@@ -2,7 +2,7 @@ package PDF::Tiny;
 
 use 5.01;
 
-$VERSION = '0.04'; # Update the POD, too!
+$VERSION = '0.05'; # Update the POD, too!
 
 # Fields
 sub vers () { 0 }
@@ -57,6 +57,7 @@ sub new {
  $self->[file] = $file;
  $self->[$_] = {} for xrft..objs; # This is why they must be consecutive.
  $self->[free] = [];
+ bless $self, $class;
  if (defined $file) {
   open my $fh, "<", $file or croak "Cannot open $file: $!";
   binmode $self->[fh] = $fh;
@@ -90,7 +91,7 @@ sub new {
   }
   else { $self->[size] = 1; }
  }
- bless $self, $class;
+ $self;
 }
 
 sub _parse_xref {
@@ -111,24 +112,61 @@ sub _parse_xref {
 		
 		seek $fh, $startxref, 0 
 			or croak "Cannot seek in $self->[file]: $!",1;
-		my $read = '';
-		while($read !~ /startxref/){
+		read $fh, my $read, 1024, length $read
+			 or croak "Cannot read $self->[file]: $!", 1;
+		if ($read =~ /^$S*[0-9]/o) { # cross-reference stream
+			my $obj = _read_obj($self, $startxref);
+			my $stream = $self->decode_stream($obj);
+			$trailer = $$obj[1];
+			my $hash = $$trailer[1];
+			my @widths = map $$_[1], @{$$hash{W}[1]};
+			my $width = $widths[0] + $widths[1] + $widths[2];
+			my $unpack = join '', map "H".$_*2, @widths;
+			my @indices = $$hash{Index}
+				? map $$_[1], @{$$hash{Index}[1]}
+				: (0, $$hash{Size}[1]);
+			my ($ix, $last) = splice @indices, 0, 2;
+			$last += $ix - 1;
+			while ($stream =~ s/^(.{$width})//s) {
+				my($type,$where,$gen)
+					= map hex, unpack $unpack, $1;
+				$widths[0] or $type = 1;
+
+				if ($type == 1) {
+					my $obj_ref = "$ix $gen";
+					!exists $$xref{$obj_ref}
+					 and $$xref{$obj_ref} = $where;
+				}
+				elsif ($type == 2) {
+					my $obj_ref = "$ix 0";
+					!exists $$xref{$obj_ref}
+					 and $$xref{$obj_ref} =
+					      ["$where 0", $gen];
+				}
+				else { # free
+					push @{$self->[free]}, "$ix $gen"
+					 if $ix && $gen != 65535
+				}
+				if ($ix++ > $last) {
+					($ix, $last) = splice @indices,0,2;
+					$last += $ix - 1;
+				}
+			}
+		}
+		else {
+		    while($read !~ /startxref/){
 			read $fh, $read, 1024, length $read
 			 or croak "Cannot read $self->[file]: $!", 1;
-		}
-		$read =~ /xref(.*?)trailer(.*)/s;
-		my $xreftext =$1;
+		    }
+		    $read =~ /xref(.*?)trailer(.*)/s;
+		    my $xreftext =$1;
 
-		$trailer = parse_string("$2",qr/^startxref\z/);
-		unless ($self->[trai]) {
-			$self->[trai] = $trailer;
-			exists $$trailer[1]{Encrypt}
-				and croak "$self->[file] is encrypted", 1;
-		}
+		    $trailer = parse_string("$2",qr/^startxref\z/);
 
-		# remove initial line, and read the numbers, repeating as necessary
+		    # remove initial line, and read the numbers,
+		    # repeating as necessary
 
-		while ($xreftext=~ s/^$N?(\d+) (\d+).*?$N//o) {
+		    while ($xreftext=~ s/^$N?(\d+) (\d+).*?$N//o) {
 			for ($1..$1+$2-1) { 	
 				$xreftext =~ s/(.{20})//s; # get 20 bytes
 				my $obj_ref = "$_ " . substr($1,11,5)*1 ;
@@ -143,6 +181,12 @@ sub _parse_xref {
 					 unless substr($1,11,5) == 65535
 				}
 			}
+		    }
+		}
+		unless ($self->[trai]) {
+			$self->[trai] = $trailer;
+			exists $$trailer[1]{Encrypt}
+				and croak "$self->[file] is encrypted", 1;
 		}
 
 		$trailer = $$trailer[1];
@@ -170,7 +214,7 @@ sub _walk_pages {
 	for (0 .. $#{$$kids[1]}){
 		$kid = $$kids[1][$_];
 		push @pages, ${$self->get_obj($kid, '/Type')}[1] eq 'Pages'
-			? _walk_pages($self->get_obj($$kid[1], '/Kids'))
+			? _walk_pages($self, $kid)
 			: $kid;
 	}
 	return @pages;
@@ -257,6 +301,9 @@ sub append {
  if (!%{$self->[mods]}) {
   return;
  }
+ if ($self->[trai][1]{Type}) {
+  croak "Cannot append to files with cross-reference streams";
+ }
  open my $fh, ">>", $self->[file]
    or croak "Cannot open $self->[file] for writing: $!";
  binmode $fh;
@@ -351,6 +398,12 @@ sub print {
   @{$$id_array[1]} = (['str', time."" ^ "".rand ^ "".(0+$self)])x2;
  }
 
+ # We assume that if this points to a cross-reference stream’s dictionary
+ # then we will not be emitting that cross-reference stream.
+ delete @{ $self->[trai][1] }{qw< XRefStm Length Filter DecodeParms F
+                                  FFilter FDecodeParms DL Type Size Index
+                                  Prev W >};
+
  my @trailer = _serialize($self->[trai]);
  my %seen;
  my @ids;
@@ -450,6 +503,33 @@ sub read_obj {
   croak "No file open";
  }
  my $loc = $self->[xrft]{$id} || return $null;
+ if (ref $loc) { # handle object streams here
+  my ($strmid, $ix) = @$loc;
+  # Since we have to decompress the stream and calculate the offsets, let’s
+  # go ahead and extract all the objects into the objects hash,  in flat
+  # format.  We may have reached this code because somebody  manually
+  # deleted an objects entry in order to re-read it, so only extract
+  # objects that are not already in the hash.
+  my $obj = $self->get_obj($strmid);
+  my $stream = \$self->decode_stream($obj);
+  my $count = $self->get_obj($$obj[1], "/N")->[1];
+  my $first = $self->get_obj($$obj[1], "/First")->[1];
+  my @numbers =
+   split /(?:$S++|%[^\cm\cj]*[\cm\cj])+/, substr $$stream, 0, $first, '';
+  while (@numbers) {
+   my ($id, $off) = splice @numbers, 0, 2;
+   $id .= " 0";
+   $self->[objs]{$id} ||=
+    ['flat',
+      substr $$stream, $off, @numbers ? $numbers[1]-$off : length $$stream]
+  }
+  return $self->[objs]{$id}
+ }
+ # otherwise use the seek-and-read approach
+ _read_obj($self, $loc, $id);
+}
+sub _read_obj {
+ my ($self, $loc, $id) = @_;
  seek $self->[fh], $loc, 0;
  read $self->[fh], my $buf, 1024 or croak "Cannot read $self->[file]: $!";
 
@@ -459,7 +539,7 @@ sub read_obj {
                          or croak "Cannot read $self->[file]: $!"
                        });
  my $read_id = 0+shift(@tokens) . ' ' . (0+shift@tokens);
- if ($read_id ne $id) {
+ if ($id and $read_id ne $id) {
   croak "$self->[file]: Found $read_id at offset $loc instead of $id";
  }
  shift @tokens; # remove “obj”
@@ -488,7 +568,7 @@ sub read_obj {
   pop @tokens; # remove ‘endobj’
   $obj = ['tokens', \@tokens];
  }
- $self->[objs]{$id} = $obj
+ $self->[objs]{$read_id} = $obj
 }
 
 sub get_obj {
@@ -533,6 +613,7 @@ sub _get_obj {
  }
  $obj or return;
  while (@_) {
+  if ($$obj[0] eq 'stream') { $obj = $$obj[1] } # for get_obj($stream,$key)
   _unflatten($obj);
   my $key = shift;
   my $subobj = $key =~ m.^/. ? $$obj[1]{substr $key, 1} : $$obj[1][$key];
@@ -552,12 +633,13 @@ sub _get_obj {
   }
  }
  _unflatten($obj);
+ $obj->[0] eq 'null' and return;
  $obj, $lastref;
 }
 sub _unflatten {
   my $obj = shift;
   if ($$obj[0] eq 'flat') {
-   $$obj[1] = _interpret_token(tokenize($$obj[1]));
+   @$obj = @{ _interpret_token([tokenize($$obj[1])]) };
   }
   elsif($$obj[0] eq 'tokens') {
    @$obj = @{ _interpret_token($$obj[1]) };
@@ -571,7 +653,7 @@ sub _viv {
                : $type =~ /^(?:array|tokens)\z/ ? []
                : $type eq 'num'                 ? 0
                : $type eq 'null'                ? ()
-               : $type eq 'stream'              ? ({}, '') : '')
+               : $type eq 'stream'              ? (['dict',{}], '') : '')
      ];
 }
 
@@ -677,6 +759,99 @@ sub add_obj {
  $id;
 }
 
+sub decode_stream :lvalue{
+ my $self = shift;
+ my $stream = $self->get_obj(@_);
+ my @filters = $self->get_obj($stream, "/Filter");
+ if ($filters[0][0] eq 'array') {
+  @filters = map $self->get_obj($filters[0],$_)->[1],0..$#{$filters[0][1]};
+ }
+ else { @filters = $filters[0][1] }
+ my @params = $self->get_obj($stream, "/DecodeParms")
+           || $self->get_obj($stream, "/DP"); # unofficial but Acrobat sup-
+ if (@params) {                               # ports it
+  if ($params[0][0] eq 'array') {
+   @params = map scalar $self->get_obj($params[0], $_),
+                 0..$#{$params[0][1]};
+  }
+ }
+ $stream = \$stream->[2];
+ for (@filters) {
+  $stream = _unfilter($self, $stream, $_, shift @params);
+ }
+ $$stream
+}
+
+sub _unfilter {
+ my ($self, $stream, $filter, $params) = @_;
+ $filter eq 'FlateDecode'
+   or croak "The $filter filter is not supported", 1;
+ my ($predictor, $bpc, $cols, $colours) = (1, 8, 1, 1);
+ if ($params) {
+  $params->[1]{Predictor}
+   and $predictor = $self->get_obj($params, "/Predictor")->[1];
+  $predictor == 1 || $predictor >= 10
+   || croak "Predictor functions other than PNG are not supported", 1;
+  $params->[1]{BitsPerComponent}
+   and $bpc = $self->get_obj($params, "/BitsPerComponent")->[1];
+  $$params[1]{Columns} and $cols=$self->get_obj($params, "/Columns")->[1];
+  $$params[1]{Colours} and $colours=$self->get_obj($params,"/Colors")->[1];
+  $bpc % 8 and croak "BitsPerComponent values that are not multiples of"
+                   . " 8 are not supported", 1;
+  $bpc >>= 3; # bytes per component
+  $bpc *= $colours;
+ }
+ require Compress::Zlib;
+ my $x = Compress'Zlib'inflateInit()
+  or croak "Could not create an inflation stream (whatever that is)", 1;
+ my ($flate_output, $flate_stat) = inflate $x my $copy = $$stream;
+ croak "Inflation failed for some reason", 1
+  unless $flate_stat == &Compress'Zlib'Z_STREAM_END;
+ if ($predictor >= 10) { # rats
+  my $output = '';
+  my $rowsize = 1 + $bpc * $cols;
+  my $prev = "\0"x($rowsize-1);
+  for my $row (1..length($flate_output) / $rowsize) {
+   my $filter = vec $flate_output, ($row-1) * $rowsize, 8;
+   my $samples = substr $flate_output, ($row-1) * $rowsize + 1, $rowsize-1;
+   if ($filter == 2) { # Up (first ’cos it’s the most common)
+    for (0..$rowsize-2) {
+     vec ($samples, $_, 8) += vec $prev, $_, 8;
+    }
+   }
+   elsif (!$filter) { # Nothing
+   }
+   elsif ($filter == 1) { # Sub (left)
+    for (0..$rowsize-2) {
+     vec ($samples, $_, 8) += vec $samples, $_ - $bpc, 8;
+    }
+   }
+   elsif ($filter == 3) { # Avg
+    for (0..$rowsize-2) {
+     vec ($samples, $_, 8) +=
+      (vec($prev, $_, 8) + vec $samples, $_ - $bpc, 8) / 2;
+    }
+   }
+   elsif ($filter == 4) { # Paeth
+    for (0..$rowsize-2) {
+     my ($a,$b,$c) = (vec($samples, $_ - $bpc, 8),
+                      vec($prev   , $_       , 8),
+                      vec $prev   , $_ - $bpc, 8 );
+     my $p = $a + $b - $c;
+     my ($pa, $pb, $pc) = (abs($p - $a), abs($p - $b), abs($p - $c));
+     vec $samples, $_, 8 =>=
+       $pa <= $pb && $pa <= $pc ? $a : $pb <= $pc ? $b : $c
+    }
+   }
+   else { croak "Invalid PNG filter value: $filter", 1 }
+   $output .= $prev = $samples;
+  }
+  \$output;
+ }
+ else {
+  \$flate_output;
+ }
+}
 
 
 # FUNCTIONS
@@ -713,7 +888,7 @@ sub tokenize { # This function tokenizes.
 	 		push @tokens, $1;
 		}
 		if (s.^\(..) {  # remove paren.
-			$more and &$more()
+			&$more()
 				until s/(
 				 (?:\\.|[^()\\])++# escaped char or non-\()
 				  |
@@ -803,6 +978,16 @@ sub _interpret_token { # pass an array ref
 		return $to_return;
 	}
 
+	# names
+
+	elsif (s.^/..) { # if it begins with "/"
+	    for (my $copy = $_) {
+		# replace #XX sequences with real chars:
+		s/#([a-f\d]{2})/chr hex $1/gie;
+		return ['name', $_];
+	    }
+	}
+
 	# dictionaries:
 
 	elsif ($_ eq '<<') {
@@ -817,6 +1002,22 @@ sub _interpret_token { # pass an array ref
 				$name =~ s/#([a-f\d]{2})/chr hex $1/gie;
 				$tmp_hash{$name} = 
 					_interpret_token($tokens);
+				delete $tmp_hash{$name}
+					if $tmp_hash{$name}[0] eq 'null'
+			}
+		}
+	}
+
+	# arrays:
+
+	elsif ($_ eq '[') {
+		my @tmp_array;
+		while(scalar @$tokens){
+			if ($$tokens[0] eq ']') {
+				shift @$tokens; #shift off the "]"
+				return ['array', \@tmp_array];
+			}else {
+				push @tmp_array, _interpret_token($tokens);
 			}
 		}
 	}
@@ -846,15 +1047,6 @@ sub _interpret_token { # pass an array ref
 		return ['str', $_];
 	}
 
-	# booleans:
-
-	elsif ($_ eq 'true') {
-		return ['bool', 1];
-	}
-	elsif($_ eq 'false'){
-		return ['bool',''];
-	}
-
 	# numbers:
 
 	elsif (/^[+\-]?\d+$/ or
@@ -870,28 +1062,13 @@ sub _interpret_token { # pass an array ref
 		return ['str', pack "H*", $_];
 	}
 
-	# names
+	# booleans:
 
-	elsif (s.^/..) { # if it begins with "/"
-	    for (my $copy = $_) {
-		# replace #XX sequences with real chars:
-		s/#([a-f\d]{2})/chr hex $1/gie;
-		return ['name', $_];
-	    }
+	elsif ($_ eq 'true') {
+		return ['bool', 1];
 	}
-
-	# arrays:
-
-	elsif ($_ eq '[') {
-		my @tmp_array;
-		while(scalar @$tokens){
-			if ($$tokens[0] eq ']') {
-				shift @$tokens; #shift off the "]"
-				return ['array', \@tmp_array];
-			}else {
-				push @tmp_array, _interpret_token($tokens);
-			}
-		}
+	elsif($_ eq 'false'){
+		return ['bool',''];
 	}
 
 	# null:
